package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.core.BlockPos;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.DoubleTag;
import net.minecraft.nbt.StringTag;
import net.minecraft.nbt.Tag;
import net.minecraft.util.StringRepresentable;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.entity.BlockEntityType;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraft.world.level.block.state.properties.EnumProperty;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;

abstract public class BlockEntityMatteryPoweredWorker extends BlockEntityMatteryPowered {
	public BlockEntityMatteryPoweredWorker(BlockEntityType<?> p_155228_, BlockPos p_155229_, BlockState p_155230_) {
		super(p_155228_, p_155229_, p_155230_);
	}

	@Nonnull
	protected abstract BigDecimal getBaseConsumption();

	public record WorkTickContext(MachineJob job, BigDecimal required_power, BigDecimal extracted_power, BigDecimal work_speed) {

	}

	public record MachineJobStatus(boolean valid, int throttle) {
		public MachineJobStatus() {
			this(true, 0);
		}

		public MachineJobStatus(boolean finished) {
			this(finished, 0);
		}
	}

	public record MachineJob(ItemStack stack, double ticks_processing_time, BigDecimal power_consumption_multiplier, CompoundTag data) {
		public MachineJob(ItemStack stack, double ticks_processing_time, BigDecimal power_consumption_multiplier, CompoundTag data) {
			this.stack = stack;
			this.ticks_processing_time = ticks_processing_time;
			this.power_consumption_multiplier = power_consumption_multiplier;
			this.data = data;
		}

		public MachineJob(ItemStack stack, double ticks_processing_time, BigDecimal power_consumption_multiplier) {
			this(stack, ticks_processing_time, power_consumption_multiplier, new CompoundTag());
		}

		public MachineJob(ItemStack stack, double ticks_processing_time) {
			this(stack, ticks_processing_time, BigDecimal.ONE);
		}

		public CompoundTag serializeNBT() {
			CompoundTag store_job = new CompoundTag();

			store_job.put("data", data);
			store_job.put("stack", stack.serializeNBT());
			store_job.putDouble("ticks_processing_time", ticks_processing_time);
			store_job.putString("power_consumption_multiplier", power_consumption_multiplier.toString());

			return store_job;
		}

		@Nullable
		public static MachineJob deserializeNBT(@Nullable Tag nbt) {
			if (nbt == null)
				return null;

			if (nbt instanceof CompoundTag tag) {
				if (tag.get("stack") instanceof CompoundTag stack_tag && tag.get("ticks_processing_time") instanceof DoubleTag ticks_processing_time && tag.get("power_consumption_multiplier") instanceof StringTag power_consumption_multiplier) {
					ItemStack stack = ItemStack.of(stack_tag);

					if (!stack.isEmpty()) {
						return new MachineJob(stack, ticks_processing_time.getAsDouble(), new BigDecimal(power_consumption_multiplier.getAsString()), tag.getCompound("data"));
					}
				}
			}

			return null;
		}
	}

	@Nonnull
	protected BigDecimal consumptionPerWork() {
		return getBaseConsumption();
	}

	protected double work_ticks = 0;
	protected int throttle_ticks = 0;
	protected MachineJob current_job;

	// if is_idling is true, then workerLoop() does nothing
	protected boolean is_idling = false;

	@Nullable
	public MachineJob getCurrentJob() {
		return current_job;
	}

	public boolean cantProcessJob() {
		return throttle_ticks > 0;
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.putDouble("work_ticks", work_ticks);

		if (current_job != null) {
			nbt.put("current_job", current_job.serializeNBT());
		}

		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);

		if (nbt.get("work_ticks") instanceof DoubleTag tag)
			work_ticks = tag.getAsDouble();

		current_job = MachineJob.deserializeNBT(nbt.get("current_job"));

		if (current_job == null)
			work_ticks = 0d;
	}

	@Override
	public void setChanged() {
		super.setChanged();
		is_idling = false;
	}

	/**
	 * @param job current job
	 * @return whenever machine can finish it's job. return false if machine for whatever reason can't finish it's job,
	 *         waiting on conditions to be met
	 */
	@Nonnull
	abstract protected MachineJobStatus onJobFinish(MachineJob job);

	/**
	 * @param context context for current job
	 * @return whenever machine can perform it
	 */
	protected MachineJobStatus onWorkTick(WorkTickContext context) {
		return new MachineJobStatus();
	}

	public enum WorkerState implements StringRepresentable {
		IDLE,
		WORKING,
		ERROR;

		@Override
		@Nonnull
		public String getSerializedName() {
			return this == IDLE ? "idle" : this == WORKING ? "working" : "error";
		}
	}

	public static final EnumProperty<WorkerState> WORKER_STATE = EnumProperty.create("worker", WorkerState.class);

	private int idle_ticks_anim = 0;
	private int working_ticks_anim = 0;

	protected void workerLoop() {
		if (throttle_ticks > 0) {
			if (level != null && throttle_ticks > 40 && getBlockState().hasProperty(WORKER_STATE) && getBlockState().getValue(WORKER_STATE) != WorkerState.ERROR) {
				level.setBlock(getBlockPos(), getBlockState().setValue(WORKER_STATE, WorkerState.ERROR), Block.UPDATE_CLIENTS);
			}

			working_ticks_anim = 0;
			idle_ticks_anim = 0;
			throttle_ticks--;

			if (throttle_ticks > 0)
				return;
		}

		if (is_idling) {
			working_ticks_anim = 0;
			idle_ticks_anim++;

			if (level != null && idle_ticks_anim > 20 && getBlockState().hasProperty(WORKER_STATE) && getBlockState().getValue(WORKER_STATE) != WorkerState.IDLE) {
				level.setBlock(getBlockPos(), getBlockState().setValue(WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS);
			}

			return;
		}

		if (current_job == null) {
			MachineJob input = getNextJob();

			if (input == null) {
				is_idling = true;
				working_ticks_anim = 0;
				return;
			}

			current_job = input;
		}

		if (current_job.power_consumption_multiplier.compareTo(BigDecimal.ZERO) != 0 && energy.getBatteryLevel().compareTo(BigDecimal.ZERO) == 0) {
			idle_ticks_anim++;

			if (level != null && idle_ticks_anim > 20 && getBlockState().hasProperty(WORKER_STATE) && getBlockState().getValue(WORKER_STATE) != WorkerState.IDLE) {
				level.setBlock(getBlockPos(), getBlockState().setValue(WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS);
			}

			return;
		}

		idle_ticks_anim = 0;

		if (work_ticks < current_job.ticks_processing_time) {
			if (current_job.power_consumption_multiplier.compareTo(BigDecimal.ZERO) != 0) {
				BigDecimal required_power = consumptionPerWork().multiply(current_job.power_consumption_multiplier);
				BigDecimal extracted_power = energy.extractEnergyInner(required_power, true);

				BigDecimal work_speed = extracted_power.divide(required_power, MatteryCapability.ROUND_RULES);

				MachineJobStatus status = onWorkTick(new WorkTickContext(current_job, required_power, extracted_power, work_speed));

				if (!status.valid) {
					throttle_ticks += status.throttle;
					return;
				}

				working_ticks_anim++;

				double new_work_ticks = work_speed.doubleValue() + work_ticks;

				if (new_work_ticks > current_job.ticks_processing_time) {
					work_ticks = current_job.ticks_processing_time;
					energy.extractEnergyInner(extracted_power.multiply(new BigDecimal(1d - (new_work_ticks - current_job.ticks_processing_time), MatteryCapability.ROUND_RULES)), false);
				} else {
					work_ticks = new_work_ticks;
					energy.extractEnergyInner(extracted_power, false);

					if (work_ticks >= current_job.ticks_processing_time) {
						MachineJobStatus finish = onJobFinish(current_job);

						if (finish.valid) {
							current_job = null;
							work_ticks = 0d;
						} else {
							throttle_ticks += finish.throttle;
						}
					}
				}
			} else {
				MachineJobStatus status = onWorkTick(new WorkTickContext(current_job, BigDecimal.ZERO, BigDecimal.ZERO, BigDecimal.ONE));

				if (!status.valid) {
					throttle_ticks += status.throttle;
					return;
				}

				working_ticks_anim++;
				work_ticks += 1d;

				if (work_ticks >= current_job.ticks_processing_time) {
					MachineJobStatus finish = onJobFinish(current_job);

					if (finish.valid) {
						current_job = null;
						work_ticks = 0d;
					} else {
						throttle_ticks += finish.throttle;
					}
				}
			}
		} else {
			MachineJobStatus finish = onJobFinish(current_job);

			if (finish.valid) {
				current_job = null;
				work_ticks = 0d;
			} else {
				throttle_ticks += finish.throttle;
			}
		}

		if (level != null && working_ticks_anim > 20 && getBlockState().hasProperty(WORKER_STATE) && getBlockState().getValue(WORKER_STATE) != WorkerState.WORKING) {
			level.setBlock(getBlockPos(), getBlockState().setValue(WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS);
		}
	}

	public double getWorkTicks() {
		return work_ticks;
	}

	public double getWorkProgress() {
		if (current_job == null)
			return 0d;

		return Math.min(1d, work_ticks / current_job.ticks_processing_time);
	}

	/**
	 * Determine which item can be processed from input slots if idling
	 * @return any item in input slots. null if no work is available
	 */
	@Nullable
	protected abstract MachineJob getNextJob();

	public static <T extends BlockEntity> void basicTicker(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityMatteryPoweredWorker tile) {
			tile.batteryChargeLoop();
			tile.workerLoop();
		}
	}
}
